-------------------------------------------------------------------------------
--
-- Copyright (C) 2019
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
-- . Reinier van der Walle
-- Purpose:
-- . Provide 10G ethernet I/O interface (BSP) for OpenCL kernel on Arria10
-- Description:
-- . This core consists of:
--   . An Intel/Altera 10G Low Latency MAC instance
--   . SOP/EOP insertion (kernel channel only carries data and valid)
--   . Dual clock FIFO 
--     . Clock domain transition between kernel_clk and clk_txmac
--     . Buffers full Ethernet packet (10G MAC requires uninterrupted packet)
--   . Clock (PLL) / reset generation
-- . Details:
--   . This core was developed for use on the Uniboard2b.
--   .
--   . The data field of the ST-avalon interface is also used to provide
--   . SOP, EOP and empty meta-data. The implementation of this is shown below.
--   +-----------+---------+--------------------------------------------------------+
--   | Bit range | Name    | Description                                            |
--   +-----------+---------+--------------------------------------------------------+
--   | [0:63]    | payload | Packet payload                                         |
--   +-----------+---------+--------------------------------------------------------+
--   | 64        | sop     | Start of packet signal                                 |
--   +-----------+---------+--------------------------------------------------------+
--   | 65        | eop     | End of packet signal                                   |
--   +-----------+---------+--------------------------------------------------------+
--   | 66:68     | -       | reserved bits                                          |
--   +-----------+---------+--------------------------------------------------------+
--   | 69:71     | empty   | On EOP, this field indicates how many bytes are unused |
--   +-----------+---------+--------------------------------------------------------+
LIBRARY IEEE, common_lib, dp_lib, tech_pll_lib, technology_lib, tech_eth_10g_lib, tech_mac_10g_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE tech_mac_10g_lib.tech_mac_10g_component_pkg.ALL;

ENTITY ta2_unb2b_10GbE IS       
  PORT (      
    --config_clk       : IN  STD_LOGIC; -- 100MHz clk for reconfig block and status interface
    config_reset     : IN  STD_LOGIC;

    clk_ref_r        : IN  STD_LOGIC; -- 644.53125MHz 10G MAC reference clock

    tx_serial_r      : OUT STD_LOGIC; -- Serial TX lanes towards QSFP cage
    rx_serial_r      : IN  STD_LOGIC; -- Serial RX lanes from QSFP cage

    kernel_clk       : IN  STD_LOGIC; -- Kernel clock (runs the kernel_* I/O below)
    kernel_reset     : IN  STD_LOGIC;

    kernel_src_data  : OUT STD_LOGIC_VECTOR(71 DOWNTO 0); -- RX Data to kernel
    kernel_src_valid : OUT STD_LOGIC; -- RX data valid signal to kernel
    kernel_src_ready : IN  STD_LOGIC; -- Flow control from kernel

    kernel_snk_data  : IN  STD_LOGIC_VECTOR(71 DOWNTO 0); -- TX Data from kernel
    kernel_snk_valid : IN  STD_LOGIC; -- TX data valid signal from kernel
    kernel_snk_ready : OUT STD_LOGIC;  -- Flow control towards kernel

    rx_status        : OUT STD_LOGIC -- RX status
  );
END ta2_unb2b_10GbE;


ARCHITECTURE str OF ta2_unb2b_10GbE IS

  CONSTANT c_sim                           : BOOLEAN := FALSE;

  CONSTANT c_nof_streams_qsfp              : NATURAL := 1;

  CONSTANT c_tx_fifo_fill                  : NATURAL := 1125; -- Largest frame is 9000 bytes = 1125
  CONSTANT c_tx_fifo_size                  : NATURAL := 2048; 
  CONSTANT c_rx_fifo_size                  : NATURAL := 256; -- should be large enough

  SIGNAL tr_ref_clk_312                    : STD_LOGIC;
  SIGNAL tr_ref_clk_156                    : STD_LOGIC;
  SIGNAL tr_ref_rst_156                    : STD_LOGIC;
  
  SIGNAL unb2_board_front_io_serial_tx_arr : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL unb2_board_front_io_serial_rx_arr : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0);
  
  SIGNAL eth_ref_clk_644                   : STD_LOGIC;
  SIGNAL eth_ref_clk_312                   : STD_LOGIC;
  SIGNAL eth_ref_clk_156                   : STD_LOGIC;
  SIGNAL eth_ref_rst_156                   : STD_LOGIC;
  
  SIGNAL eth_tx_clk_arr                    : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0);
  SIGNAL eth_tx_rst_arr                    : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0);
  SIGNAL eth_rx_clk_arr                    : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0);
  SIGNAL eth_rx_rst_arr                    : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0);

  SIGNAL dp_latency_adapter_tx_src_out     : t_dp_sosi; 
  SIGNAL dp_latency_adapter_tx_src_in      : t_dp_siso;

  SIGNAL dp_latency_adapter_tx_snk_in      : t_dp_sosi; 
  SIGNAL dp_latency_adapter_tx_snk_out     : t_dp_siso;

  SIGNAL dp_fifo_fill_tx_src_out_arr       : t_dp_sosi_arr(c_nof_streams_qsfp-1 DOWNTO 0); 
  SIGNAL dp_fifo_fill_tx_src_in_arr        : t_dp_siso_arr(c_nof_streams_qsfp-1 DOWNTO 0);

  SIGNAL mac_10g_src_out_arr               : t_dp_sosi_arr(c_nof_streams_qsfp-1 DOWNTO 0); 
  SIGNAL mac_10g_src_in_arr                : t_dp_siso_arr(c_nof_streams_qsfp-1 DOWNTO 0);

  SIGNAL dp_fifo_dc_rx_src_out             : t_dp_sosi; 
  SIGNAL dp_fifo_dc_rx_src_in              : t_dp_siso;

  SIGNAL dp_latency_adapter_rx_src_out     : t_dp_sosi; 
  SIGNAL dp_latency_adapter_rx_src_in      : t_dp_siso;

  SIGNAL dp_xonoff_src_out                 : t_dp_sosi;
  SIGNAL dp_xonoff_src_in                  : t_dp_siso;
BEGIN

  --------
  -- PLL
  --------
  u_tech_pll_xgmii_mac_clocks : ENTITY tech_pll_lib.tech_pll_xgmii_mac_clocks
  GENERIC MAP (
    g_technology => c_tech_arria10_e1sg
  )
  PORT MAP (
    refclk_644 => clk_ref_r,
    rst_in     => config_reset,
    clk_156    => tr_ref_clk_156,
    clk_312    => tr_ref_clk_312,
    rst_156    => tr_ref_rst_156,
    rst_312    => OPEN
  );


  ---------------------------------------------------------------------------------------
  -- Clocks and reset
  ---------------------------------------------------------------------------------------

  -- Apply the clocks from top level down such that they have their rising_edge() aligned without any delta-delay
  u_tech_eth_10g_clocks : ENTITY tech_eth_10g_lib.tech_eth_10g_clocks
  GENERIC MAP (
    g_technology     => c_tech_arria10_e1sg,
    g_nof_channels   => c_nof_streams_qsfp
  )
  PORT MAP (
    -- Input clocks
    -- . Reference
    tr_ref_clk_644    => clk_ref_r,
    tr_ref_clk_312    => tr_ref_clk_312,
    tr_ref_clk_156    => tr_ref_clk_156,
    tr_ref_rst_156    => tr_ref_rst_156,
    
    
    -- Output clocks
    -- . Reference
    eth_ref_clk_644   => eth_ref_clk_644,
    eth_ref_clk_312   => eth_ref_clk_312,
    eth_ref_clk_156   => eth_ref_clk_156,
    eth_ref_rst_156   => eth_ref_rst_156,
    
    -- . Data
    eth_tx_clk_arr    => eth_tx_clk_arr,
    eth_tx_rst_arr    => eth_tx_rst_arr,
    
    eth_rx_clk_arr    => eth_rx_clk_arr,
    eth_rx_rst_arr    => eth_rx_rst_arr
  );  
 

  ----------------------------------------------------------------------------
  -- Data mapping 
  ----------------------------------------------------------------------------
  -- Reverse byte order
  gen_tx_bytes: FOR I IN 0 TO 7 GENERATE
    dp_latency_adapter_tx_snk_in.data(8*(8-I) -1  DOWNTO 8*(7-I)) <= kernel_snk_data(8*(I+1) -1 DOWNTO 8*I);
  END GENERATE;

  -- Assign correct data fields to control signals.
  dp_latency_adapter_tx_snk_in.sop <= kernel_snk_data(64);
  dp_latency_adapter_tx_snk_in.eop <= kernel_snk_data(65);
  dp_latency_adapter_tx_snk_in.empty(2 DOWNTO 0) <= kernel_snk_data(71 DOWNTO 69);

  dp_latency_adapter_tx_snk_in.valid <= kernel_snk_valid;
  kernel_snk_ready <= dp_latency_adapter_tx_snk_out.ready; -- Flow control towards source (kernel)

  rx_status <= dp_xonoff_src_in.xon; -- use xonoff_src_in for status as xonoff_snk_out is always '1'

  tx_serial_r <= unb2_board_front_io_serial_tx_arr(0);
  unb2_board_front_io_serial_rx_arr(0) <= rx_serial_r;

  ----------------------------------------------------------------------------
  -- Latency adapter: adapt RL=0 (OpenCL kernel) to RL=1 (Downstream).
  ----------------------------------------------------------------------------
  u_dp_latency_adapter_tx : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency  => 0,
    g_out_latency => 1 
  )
  PORT MAP (
    clk       => kernel_clk,
    rst       => kernel_reset,

    snk_in    => dp_latency_adapter_tx_snk_in, 
    snk_out   => dp_latency_adapter_tx_snk_out, 

    src_out   => dp_latency_adapter_tx_src_out, 
    src_in    => dp_latency_adapter_tx_src_in 
  );
  
    
  -----------------------------------------------------------------------------
  -- RX XON frame control
  -----------------------------------------------------------------------------

  u_dp_xonoff : ENTITY dp_lib.dp_xonoff
  PORT MAP (
    rst      => kernel_reset,
    clk      => kernel_clk,
         
    in_siso  => dp_latency_adapter_tx_src_in,      
    in_sosi  => dp_latency_adapter_tx_src_out,
    
    out_siso => dp_xonoff_src_in,
    out_sosi => dp_xonoff_src_out    
  );   
 

  ---------------------------------------------------------------------------------------
  -- FIFO FILL with fill level/eop trigger so we can deliver packets to the MAC fast enough
  ---------------------------------------------------------------------------------------

  u_dp_fifo_fill_tx_eop : ENTITY dp_lib.dp_fifo_fill_eop
  GENERIC MAP (
    g_technology     => c_tech_arria10_e1sg,
    g_use_dual_clock => TRUE,
    g_data_w         => c_xgmii_data_w,
    g_empty_w        => c_tech_mac_10g_empty_w,
    g_use_empty      => TRUE,
    g_fifo_fill      => c_tx_fifo_fill,
    g_fifo_size      => c_tx_fifo_size
  )
  PORT MAP (
    wr_rst      => kernel_reset, 
    wr_clk      => kernel_clk, 
    rd_rst      => eth_tx_rst_arr(0),
    rd_clk      => eth_tx_clk_arr(0),

    snk_out     => dp_xonoff_src_in,
    snk_in      => dp_xonoff_src_out,

    src_in      => dp_fifo_fill_tx_src_in_arr(0), 
    src_out     => dp_fifo_fill_tx_src_out_arr(0) 
  );


  ---------------------------------------------------------------------------------------
  -- ETH MAC + PHY
  ---------------------------------------------------------------------------------------

  u_tech_eth_10g : ENTITY tech_eth_10g_lib.tech_eth_10g
  GENERIC MAP (
    g_technology          => c_tech_arria10_e1sg,
    g_sim                 => c_sim,
    g_sim_level           => 1,  -- 0 = use IP; 1 = use fast serdes model
    g_nof_channels        => c_nof_streams_qsfp,
    g_direction           => "TX_RX",
    g_pre_header_padding  => FALSE
  )
  PORT MAP (
    -- Transceiver PLL reference clock
    tr_ref_clk_644   => eth_ref_clk_644,   -- 644.531250 MHz for 10GBASE-R
    tr_ref_clk_312   => eth_ref_clk_312,   -- 312.5      MHz for 10GBASE-R
    tr_ref_clk_156   => eth_ref_clk_156,   -- 156.25     MHz for 10GBASE-R or for XAUI
    tr_ref_rst_156   => eth_ref_rst_156,   --                for 10GBASE-R or for XAUI
    
    -- MM
    mm_clk           => '0',
    mm_rst           => '0',
  
    -- ST
    tx_snk_in_arr    => dp_fifo_fill_tx_src_out_arr,      -- 64 bit data @ 156 MHz
    tx_snk_out_arr   => dp_fifo_fill_tx_src_in_arr,
    
    rx_src_out_arr   => mac_10g_src_out_arr,              -- 64 bit data @ 156 MHz
    rx_src_in_arr    => mac_10g_src_in_arr,
    
    -- PHY serial IO
    -- . 10GBASE-R (single lane)
    serial_tx_arr    => unb2_board_front_io_serial_tx_arr,
    serial_rx_arr    => unb2_board_front_io_serial_rx_arr
  );



  ---------------------------------------------------------------------------------------
  -- RX FIFO: rx_clk -> dp_clk
  ---------------------------------------------------------------------------------------

  u_dp_fifo_dc_rx : ENTITY dp_lib.dp_fifo_dc
  GENERIC MAP (
    g_technology  => c_tech_arria10_e1sg,
    g_data_w      => c_xgmii_data_w,
    g_empty_w     => c_tech_mac_10g_empty_w,
    g_use_empty   => TRUE,
    g_fifo_size   => c_rx_fifo_size
  )
  PORT MAP (
    wr_rst      => eth_rx_rst_arr(0),
    wr_clk      => eth_rx_clk_arr(0),
    rd_rst      => kernel_reset,
    rd_clk      => kernel_clk,
  
    snk_out     => mac_10g_src_in_arr(0),
    snk_in      => mac_10g_src_out_arr(0),
  
    src_in      => dp_fifo_dc_rx_src_in, 
    src_out     => dp_fifo_dc_rx_src_out
  );   


  ----------------------------------------------------------------------------
  -- Latency adapter: adapt RL=1 (Upstream) to RL=0 (OpenCL kernel).
  ----------------------------------------------------------------------------
  u_dp_latency_adapter_rx : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency  => 1,
    g_out_latency => 0 
  )
  PORT MAP (
    clk       => kernel_clk,
    rst       => kernel_reset,

    snk_in    => dp_fifo_dc_rx_src_out, 
    snk_out   => dp_fifo_dc_rx_src_in, 

    src_out   => dp_latency_adapter_rx_src_out, 
    src_in    => dp_latency_adapter_rx_src_in 
  );
      

  ----------------------------------------------------------------------------
  -- Data mapping 
  ----------------------------------------------------------------------------
  -- Reverse byte order
  gen_rx_bytes: FOR I IN 0 TO 7 GENERATE
    kernel_src_data(8*(8-I) -1  DOWNTO 8*(7-I)) <= dp_latency_adapter_rx_src_out.data(8*(I+1) -1 DOWNTO 8*I);
  END GENERATE;

  -- Assign control signals to correct data fields.
  kernel_src_data(64) <= dp_latency_adapter_rx_src_out.sop;
  kernel_src_data(65) <= dp_latency_adapter_rx_src_out.eop;
  kernel_src_data(71 DOWNTO 69) <= dp_latency_adapter_rx_src_out.empty(2 DOWNTO 0);
  

  kernel_src_valid <= dp_latency_adapter_rx_src_out.valid;
  dp_latency_adapter_rx_src_in.ready <= kernel_src_ready;
  dp_latency_adapter_rx_src_in.xon <= '1';


END str;

