-------------------------------------------------------------------------------
--
-- Copyright (C) 2019
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
-- . Daniel van der Schuur
-- . Reinier van der Walle
-- Purpose:
-- . Provide 40G ethernet I/O interface (BSP) for OpenCL kernel on Arria10
-- Description:
-- . This core consists of:
--   . An Intel/Altera 40G Low Latency MAC instance
--   . SOP/EOP insertion (kernel channel only carries data and valid)
--   . Dual clock FIFO 
--     . Clock domain transition between kernel_clk and clk_txmac
--     . Buffers full Ethernet packet (40G MAC requires uninterrupted packet)
--   . Clock (PLL) / reset generation
-- . Details:
--   . This core was developed for use on the Uniboard2b.
--   .
--   . The data field of the ST-avalon interface is also used to provide
--   . SOP, EOP and empty meta-data. The implementation of this is shown below.
--   +-----------+---------+--------------------------------------------------------+
--   | Bit range | Name    | Description                                            |
--   +-----------+---------+--------------------------------------------------------+
--   | [0:255]   | payload | Packet payload                                         |
--   +-----------+---------+--------------------------------------------------------+
--   | 256       | sop     | Start of packet signal                                 |
--   +-----------+---------+--------------------------------------------------------+
--   | 257       | eop     | End of packet signal                                   |
--   +-----------+---------+--------------------------------------------------------+
--   | 258       | -       | reserved bit                                           |
--   +-----------+---------+--------------------------------------------------------+
--   | 259:263   | empty   | On EOP, this field indicates how many bytes are unused |
--   +-----------+---------+--------------------------------------------------------+
LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL; 

ENTITY ta2_unb2b_40GbE IS   
  GENERIC (
    g_nof_mac : NATURAL := 1
  );    
  PORT (      
    config_clk       : IN  STD_LOGIC; -- 100MHz clk for reconfig block and status interface
    config_reset     : IN  STD_LOGIC;

    clk_ref_r        : IN  STD_LOGIC; -- 644.53125MHz 40G MAC reference clock

    tx_serial_r      : OUT STD_LOGIC_VECTOR(4*g_nof_mac-1 DOWNTO 0); -- Serial TX lanes towards QSFP cage
    rx_serial_r      : IN  STD_LOGIC_VECTOR(4*g_nof_mac-1 DOWNTO 0); -- Serial RX lanes from QSFP cage

    kernel_clk       : IN  STD_LOGIC; -- Kernel clock (runs the kernel_* I/O below)
    kernel_reset     : IN  STD_LOGIC;

    src_out_arr      : OUT t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
    src_in_arr       : IN  t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);
    snk_out_arr      : OUT t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);
    snk_in_arr       : IN  t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0)
    
  );
END ta2_unb2b_40GbE;


ARCHITECTURE str OF ta2_unb2b_40GbE IS

  CONSTANT c_max_packet_size : NATURAL := 128; -- 128 * 256 bits words
  CONSTANT c_tx_fifo_fill    : NATURAL := 282; -- Largest frame is 9000 bytes = 1125
  CONSTANT c_tx_fifo_size    : NATURAL := 512; 
  CONSTANT c_rx_fifo_size    : NATURAL := 64; -- should be large enough


  CONSTANT c_data_w : NATURAL := 256;
  ----------------------------------------------------------------------------
  -- Reset signals
  ----------------------------------------------------------------------------
  SIGNAL rst_txmac_arr  : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);
  SIGNAL rst_rxmac_arr  : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- dp_xonoff 
  ----------------------------------------------------------------------------
  SIGNAL dp_xonoff_src_out_arr  : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_xonoff_src_in_arr   : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- Latency adapter tx a
  ----------------------------------------------------------------------------
  SIGNAL dp_latency_adapter_tx_a_src_out_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_latency_adapter_tx_a_src_in_arr  : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_latency_adapter_tx_a_snk_in_arr  : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_latency_adapter_tx_a_snk_out_arr : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- Latency adapter tx b
  ----------------------------------------------------------------------------
  SIGNAL dp_latency_adapter_tx_b_src_out_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_latency_adapter_tx_b_src_in_arr  : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- Latency adapter rx
  ----------------------------------------------------------------------------
  SIGNAL dp_latency_adapter_rx_src_out_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_latency_adapter_rx_src_in_arr  : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- TX FIFO
  ----------------------------------------------------------------------------
  SIGNAL dp_fifo_fill_eop_src_out_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_fifo_fill_eop_src_in_arr  : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- 40G MAC IP
  ----------------------------------------------------------------------------
  SIGNAL serial_clk_2arr : t_slv_4_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL serial_clk_arr : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);

  SIGNAL pll_locked_arr : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);

  SIGNAL rx_pcs_ready_arr : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);

  SIGNAL clk_txmac_arr : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);    -- MAC + PCS clock - at least 312.5Mhz
  SIGNAL clk_rxmac_arr : STD_LOGIC_VECTOR(g_nof_mac-1 DOWNTO 0);    -- MAC + PCS clock - at least 312.5Mhz

  SIGNAL l4_tx_sosi_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL l4_tx_siso_arr : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL l4_rx_sosi_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL l4_rx_siso_arr : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- RX FIFO
  ----------------------------------------------------------------------------
  SIGNAL dp_fifo_dc_src_out_arr : t_dp_sosi_arr(g_nof_mac-1 DOWNTO 0);
  SIGNAL dp_fifo_dc_src_in_arr  : t_dp_siso_arr(g_nof_mac-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- ATX PLL Component
  ----------------------------------------------------------------------------
  COMPONENT arria10_40g_atx_pll IS
  PORT (
    pll_cal_busy  : OUT STD_LOGIC;        -- pll_cal_busy
    pll_locked    : OUT STD_LOGIC;        -- pll_locked
    pll_powerdown : IN  STD_LOGIC := 'X'; -- pll_powerdown
    pll_refclk0   : IN  STD_LOGIC := 'X'; -- clk
    tx_serial_clk : OUT STD_LOGIC         -- clk
  );
  END COMPONENT arria10_40g_atx_pll;

  ----------------------------------------------------------------------------
  -- 40G ETH IP Component
  ----------------------------------------------------------------------------
  COMPONENT arria10_40g_mac IS
  PORT (
    l4_rx_error           : OUT STD_LOGIC_VECTOR(5 DOWNTO 0);                      -- l4_rx_error
    l4_rx_status          : OUT STD_LOGIC_VECTOR(2 DOWNTO 0);                      -- l4_rx_status
    l4_rx_valid           : OUT STD_LOGIC;                                         -- l4_rx_valid
    l4_rx_startofpacket   : OUT STD_LOGIC;                                         -- l4_rx_startofpacket
    l4_rx_endofpacket     : OUT STD_LOGIC;                                         -- l4_rx_endofpacket
    l4_rx_data            : OUT STD_LOGIC_VECTOR(255 DOWNTO 0);                    -- l4_rx_data
    l4_rx_empty           : OUT STD_LOGIC_VECTOR(4 DOWNTO 0);                      -- l4_rx_empty
    l4_rx_fcs_error       : OUT STD_LOGIC;                                         -- l4_rx_fcs_error
    l4_rx_fcs_valid       : OUT STD_LOGIC;                                         -- l4_rx_fcs_valid
    l4_tx_startofpacket   : IN  STD_LOGIC                      := 'X';             -- l4_tx_startofpacket
    l4_tx_endofpacket     : IN  STD_LOGIC                      := 'X';             -- l4_tx_endofpacket
    l4_tx_valid           : IN  STD_LOGIC                      := 'X';             -- l4_tx_valid
    l4_tx_ready           : OUT STD_LOGIC;                                         -- l4_tx_ready
    l4_tx_empty           : IN  STD_LOGIC_VECTOR(4 DOWNTO 0)   := (OTHERS => 'X'); -- l4_tx_empty
    l4_tx_data            : IN  STD_LOGIC_VECTOR(255 DOWNTO 0) := (OTHERS => 'X'); -- l4_tx_data
    l4_tx_error           : IN  STD_LOGIC                      := 'X';             -- l4_tx_error
    clk_ref               : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- clk_ref
    clk_rxmac             : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- clk_rxmac
    clk_status            : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- clk_status
    clk_txmac             : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- clk_txmac
    reconfig_address      : IN  STD_LOGIC_VECTOR(11 DOWNTO 0)  := (OTHERS => 'X'); -- reconfig_address
    reconfig_clk          : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- reconfig_clk
    reconfig_read         : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- reconfig_read
    reconfig_readdata     : OUT STD_LOGIC_VECTOR(31 DOWNTO 0);                     -- reconfig_readdata
    reconfig_reset        : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- reconfig_reset
    reconfig_waitrequest  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- reconfig_waitrequest
    reconfig_write        : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- reconfig_write
    reconfig_writedata    : IN  STD_LOGIC_VECTOR(31 DOWNTO 0)  := (OTHERS => 'X'); -- reconfig_writedata
    reset_async           : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- reset_async
    reset_status          : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- reset_status
    rx_pcs_ready          : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_pcs_ready
    rx_serial             : IN  STD_LOGIC_VECTOR(3 DOWNTO 0)   := (OTHERS => 'X'); -- rx_serial
    rx_inc_octetsOK       : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);                     -- rx_inc_octetsOK
    rx_inc_octetsOK_valid : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_octetsOK_valid
    rx_inc_runt           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_runt
    rx_inc_64             : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_64
    rx_inc_127            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_127
    rx_inc_255            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_255
    rx_inc_511            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_511
    rx_inc_1023           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_1023
    rx_inc_1518           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_1518
    rx_inc_max            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_max
    rx_inc_over           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_over
    rx_inc_mcast_data_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_mcast_data_err
    rx_inc_mcast_data_ok  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_mcast_data_ok
    rx_inc_bcast_data_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_bcast_data_err
    rx_inc_bcast_data_ok  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_bcast_data_ok
    rx_inc_ucast_data_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_ucast_data_err
    rx_inc_ucast_data_ok  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_ucast_data_ok
    rx_inc_mcast_ctrl     : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_mcast_ctrl
    rx_inc_bcast_ctrl     : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_bcast_ctrl
    rx_inc_ucast_ctrl     : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_ucast_ctrl
    rx_inc_pause          : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_pause
    rx_inc_fcs_err        : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_fcs_err
    rx_inc_fragment       : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_fragment
    rx_inc_jabber         : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_jabber
    rx_inc_sizeok_fcserr  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_sizeok_fcserr
    rx_inc_pause_ctrl_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_pause_ctrl_err
    rx_inc_mcast_ctrl_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_mcast_ctrl_err
    rx_inc_bcast_ctrl_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_bcast_ctrl_err
    rx_inc_ucast_ctrl_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- rx_inc_ucast_ctrl_err
    status_write          : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- status_write
    status_read           : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- status_read
    status_addr           : IN  STD_LOGIC_VECTOR(15 DOWNTO 0)  := (OTHERS => 'X'); -- status_addr
    status_writedata      : IN  STD_LOGIC_VECTOR(31 DOWNTO 0)  := (OTHERS => 'X'); -- status_writedata
    status_readdata       : OUT STD_LOGIC_VECTOR(31 DOWNTO 0);                     -- status_readdata
    status_readdata_valid : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- status_readdata_valid
    status_waitrequest    : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- status_waitrequest
    status_read_timeout   : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- status_read_timeout
    tx_lanes_stable       : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_lanes_stable
    tx_pll_locked         : IN  STD_LOGIC_VECTOR(0 DOWNTO 0)   := (OTHERS => 'X'); -- tx_pll_locked
    tx_serial             : OUT STD_LOGIC_VECTOR(3 DOWNTO 0);                      -- tx_serial
    tx_serial_clk         : IN  STD_LOGIC_VECTOR(3 DOWNTO 0)   := (OTHERS => 'X'); -- tx_serial_clk
    tx_inc_octetsOK       : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);                     -- tx_inc_octetsOK
    tx_inc_octetsOK_valid : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_octetsOK_valid
    tx_inc_64             : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_64
    tx_inc_127            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_127
    tx_inc_255            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_255
    tx_inc_511            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_511
    tx_inc_1023           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_1023
    tx_inc_1518           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_1518
    tx_inc_max            : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_max
    tx_inc_over           : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_over
    tx_inc_mcast_data_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_mcast_data_err
    tx_inc_mcast_data_ok  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_mcast_data_ok
    tx_inc_bcast_data_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_bcast_data_err
    tx_inc_bcast_data_ok  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_bcast_data_ok
    tx_inc_ucast_data_err : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_ucast_data_err
    tx_inc_ucast_data_ok  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_ucast_data_ok
    tx_inc_mcast_ctrl     : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_mcast_ctrl
    tx_inc_bcast_ctrl     : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_bcast_ctrl
    tx_inc_ucast_ctrl     : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_ucast_ctrl
    tx_inc_pause          : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_pause
    tx_inc_fcs_err        : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_fcs_err
    tx_inc_fragment       : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_fragment
    tx_inc_jabber         : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);                      -- tx_inc_jabber
    tx_inc_sizeok_fcserr  : OUT STD_LOGIC_VECTOR(0 DOWNTO 0)                       -- tx_inc_sizeok_fcserr
  );
  END COMPONENT arria10_40g_mac;



BEGIN

  gen_mac: FOR mac IN 0 TO g_nof_mac-1 GENERATE
    ----------------------------------------------------------------------------
    -- Data mapping 
    ----------------------------------------------------------------------------
    -- Reverse byte order
    gen_tx_bytes: FOR I IN 0 TO 31 GENERATE
      dp_latency_adapter_tx_a_snk_in_arr(mac).data(8*(32-I) -1  DOWNTO 8*(31-I)) <= snk_in_arr(mac).data(8*(I+1) -1 DOWNTO 8*I);
    END GENERATE;

    -- Assign correct data fields to control signals.
    dp_latency_adapter_tx_a_snk_in_arr(mac).sop <= snk_in_arr(mac).data(256);
    dp_latency_adapter_tx_a_snk_in_arr(mac).eop <= snk_in_arr(mac).data(257);
    dp_latency_adapter_tx_a_snk_in_arr(mac).empty(4 DOWNTO 0) <= snk_in_arr(mac).data(263 DOWNTO 259);
    dp_latency_adapter_tx_a_snk_in_arr(mac).valid <= snk_in_arr(mac).valid;
    snk_out_arr(mac).ready <= dp_latency_adapter_tx_a_snk_out_arr(mac).ready; -- Flow control towards source (kernel)
    snk_out_arr(mac).xon   <= rx_pcs_ready_arr(mac); 

    ----------------------------------------------------------------------------
    -- Latency adapter: adapt RL=0 (OpenCL kernel) to RL=1 (downstream).
    ----------------------------------------------------------------------------
    u_dp_latency_adapter_tx_a : ENTITY dp_lib.dp_latency_adapter
    GENERIC MAP (
      g_in_latency  => 0,
      g_out_latency => 1 
    )
    PORT MAP (
      clk       => kernel_clk,
      rst       => kernel_reset,

      snk_in    => dp_latency_adapter_tx_a_snk_in_arr(mac),
      snk_out   => dp_latency_adapter_tx_a_snk_out_arr(mac),

      src_out   => dp_latency_adapter_tx_a_src_out_arr(mac),
      src_in    => dp_latency_adapter_tx_a_src_in_arr(mac)
    );
    
    ----------------------------------------------------------------------------
    -- dp_xonoff: discard all TX frames until 40G MAC TX side is ready
    ----------------------------------------------------------------------------
    u_dp_xonoff : ENTITY dp_lib.dp_xonoff
    PORT MAP (
      clk           => kernel_clk,
      rst           => kernel_reset,

      in_sosi       => dp_latency_adapter_tx_a_src_out_arr(mac),
      in_siso       => dp_latency_adapter_tx_a_src_in_arr(mac),

      out_sosi      => dp_xonoff_src_out_arr(mac),
      out_siso      => dp_xonoff_src_in_arr(mac)  -- flush control via out_siso.xon
    );

    ----------------------------------------------------------------------------
    -- TX FIFO
    ----------------------------------------------------------------------------
    u_dp_fifo_fill_eop : ENTITY dp_lib.dp_fifo_fill_eop
    GENERIC MAP (
      g_data_w         => c_data_w,
      g_use_dual_clock => TRUE,
      g_empty_w        => 8,
      g_use_empty      => TRUE,
      g_use_bsn        => FALSE,
      g_bsn_w          => 64,
      g_use_channel    => FALSE,
      g_use_sync       => FALSE,
      g_fifo_size      => c_tx_fifo_size,
      g_fifo_fill      => c_tx_fifo_fill
   )
    PORT MAP (
      wr_clk  => kernel_clk,
      wr_rst  => kernel_reset,

      rd_clk  => clk_txmac_arr(mac),
      rd_rst  => rst_txmac_arr(mac),

      snk_in  => dp_xonoff_src_out_arr(mac),
      snk_out => dp_xonoff_src_in_arr(mac),

      src_out => dp_fifo_fill_eop_src_out_arr(mac),
      src_in  => dp_fifo_fill_eop_src_in_arr(mac)
    );

    ----------------------------------------------------------------------------
    -- Latency adapter: adapt RL=1 (upstream) to RL=0 (MAC TX interface).
    ----------------------------------------------------------------------------
    u_dp_latency_adapter_tx_b : ENTITY dp_lib.dp_latency_adapter
    GENERIC MAP (
      g_in_latency  => 1,
      g_out_latency => 0 
    )
    PORT MAP (
      clk       => clk_txmac_arr(mac),
      rst       => rst_txmac_arr(mac),

      snk_in    => dp_fifo_fill_eop_src_out_arr(mac),
      snk_out   => dp_fifo_fill_eop_src_in_arr(mac),

      src_out   => dp_latency_adapter_tx_b_src_out_arr(mac),
      src_in    => dp_latency_adapter_tx_b_src_in_arr(mac)
    );

    ----------------------------------------------------------------------------
    -- 40G MAC IP
    ----------------------------------------------------------------------------
    l4_tx_sosi_arr(mac) <= dp_latency_adapter_tx_b_src_out_arr(mac);
    dp_latency_adapter_tx_b_src_in_arr(mac) <= l4_tx_siso_arr(mac);  

    u_arria10_40g_mac : arria10_40g_mac 
      PORT MAP (
         reset_async(0)         => config_reset,
         clk_txmac(0)           => clk_txmac_arr(mac),    -- MAC + PCS clock - at least 312.5Mhz
         clk_rxmac(0)           => clk_rxmac_arr(mac),    -- MAC + PCS clock - at least 312.5Mhz
         clk_ref(0)             => clk_ref_r,
         rx_pcs_ready(0)        => rx_pcs_ready_arr(mac),

         tx_serial_clk          => serial_clk_2arr(mac),
         tx_pll_locked(0)       => pll_locked_arr(mac),

         clk_status(0)          => config_clk,
         reset_status(0)        => config_reset,
         status_addr            => (OTHERS=>'0'),
         status_read            => (OTHERS=>'0'),
         status_write           => (OTHERS=>'0'),
         status_writedata       => (OTHERS=>'0'),
--       status_readdata        => status_readdata_eth,
--       status_read_timeout    => status_read_timeout,
--       status_readdata_valid  => status_readdata_valid_eth,
      
         reconfig_clk(0)        => config_clk,
         reconfig_reset(0)      => config_reset,
         reconfig_write         => (OTHERS=>'0'),
         reconfig_read          => (OTHERS=>'0'),
         reconfig_address       => (OTHERS=>'0'),
         reconfig_writedata     => (OTHERS=>'0'),
--       reconfig_readdata      => reco_readdata[31:0],
--       reconfig_waitrequest   => reco_waitrequest,
         
         l4_tx_data             => l4_tx_sosi_arr(mac).data(255 DOWNTO 0), 
         l4_tx_empty            => l4_tx_sosi_arr(mac).empty(4 DOWNTO 0), 
         l4_tx_startofpacket    => l4_tx_sosi_arr(mac).sop, 
         l4_tx_endofpacket      => l4_tx_sosi_arr(mac).eop,
         l4_tx_ready            => l4_tx_siso_arr(mac).ready, 
         l4_tx_valid            => l4_tx_sosi_arr(mac).valid, 
         l4_tx_error            => '0',
         
         l4_rx_data             => l4_rx_sosi_arr(mac).data(255 DOWNTO 0),
         l4_rx_empty            => l4_rx_sosi_arr(mac).empty(4 DOWNTO 0),
         l4_rx_startofpacket    => l4_rx_sosi_arr(mac).sop,
         l4_rx_endofpacket      => l4_rx_sosi_arr(mac).eop,
  --       l4_rx_error            => ,
         l4_rx_valid            => l4_rx_sosi_arr(mac).valid,
  
  --        l4_rx_status                  (),
  --        l4_rx_fcs_error               (),
  --        l4_rx_fcs_valid               (),
  --        rx_inc_octetsOK               (),
  --        rx_inc_octetsOK_valid         (),
  --        rx_inc_runt                   (),
  --        rx_inc_64                     (),
  --        rx_inc_127                    (),
  --        rx_inc_255                    (),
  --        rx_inc_511                    (),
  --        rx_inc_1023                   (),
  --        rx_inc_1518                   (),
  --        rx_inc_max                    (),
  --        rx_inc_over                   (),
  --        rx_inc_mcast_data_err         (),
  --        rx_inc_mcast_data_ok          (),
  --        rx_inc_bcast_data_err         (),
  --        rx_inc_bcast_data_ok          (),
  --        rx_inc_ucast_data_err         (),
  --        rx_inc_ucast_data_ok          (),
  --        rx_inc_mcast_ctrl             (),
  --        rx_inc_bcast_ctrl             (),
  --        rx_inc_ucast_ctrl             (),
  --        rx_inc_pause                  (),
  --        rx_inc_fcs_err                (),
  --        rx_inc_fragment               (),
  --        rx_inc_jabber                 (),
  --        rx_inc_sizeok_fcserr          (),
  --        rx_inc_pause_ctrl_err         (),
  --        rx_inc_mcast_ctrl_err         (),
  --        rx_inc_bcast_ctrl_err         (),
  --        rx_inc_ucast_ctrl_err         (),
  --        status_waitrequest            (),
         tx_lanes_stable(0)     => l4_tx_siso_arr(mac).xon,
  --        tx_inc_octetsOK               (),
  --        tx_inc_octetsOK_valid         (),
  --        tx_inc_64                     (),
  --        tx_inc_127                    (),
  --        tx_inc_255                    (),
  --        tx_inc_511                    (),
  --        tx_inc_1023                   (),
  --        tx_inc_1518                   (),
  --        tx_inc_max                    (),
  --        tx_inc_over                   (),
  --        tx_inc_mcast_data_err         (),
  --        tx_inc_mcast_data_ok          (),
  --        tx_inc_bcast_data_err         (),
  --        tx_inc_bcast_data_ok          (),
  --        tx_inc_ucast_data_err         (),
  --        tx_inc_ucast_data_ok          (),
  --        tx_inc_mcast_ctrl             (),
  --        tx_inc_bcast_ctrl             (),
  --        tx_inc_ucast_ctrl             (),
  --        tx_inc_pause                  (),
  --        tx_inc_fcs_err                (),
  --        tx_inc_fragment               (),
  --        tx_inc_jabber                 (),
  --        tx_inc_sizeok_fcserr          (),

         tx_serial              => tx_serial_r(4*(mac+1)-1 DOWNTO 4*mac),
         rx_serial              => rx_serial_r(4*(mac+1)-1 DOWNTO 4*mac)
      );


    -- No latency adapter needed as the RX MAC does not have a ready input
    ----------------------------------------------------------------------------
    -- RX FIFO
    ----------------------------------------------------------------------------
    rst_rxmac_arr(mac) <= NOT rx_pcs_ready_arr(mac);
    u_dp_fifo_dc : ENTITY dp_lib.dp_fifo_dc
    GENERIC MAP (
      g_data_w         => c_data_w,
      g_empty_w        => 8,
      g_use_empty      => TRUE,
      g_use_bsn        => FALSE,
      g_bsn_w          => 64,
      g_use_channel    => FALSE,
      g_use_sync       => FALSE,
      g_fifo_size      => c_rx_fifo_size
   )
    PORT MAP (
      wr_clk  => clk_rxmac_arr(mac),
      wr_rst  => rst_rxmac_arr(mac),

      rd_clk  => kernel_clk,
      rd_rst  => kernel_reset,

      snk_in  => l4_rx_sosi_arr(mac),
      snk_out => OPEN,

      src_out => dp_fifo_dc_src_out_arr(mac),
      src_in  => dp_fifo_dc_src_in_arr(mac)
    );


    ----------------------------------------------------------------------------
    -- Latency adapter: adapt RL=1 (dp_fifo_dc) to RL=0 (OpenCL kernel).
    ----------------------------------------------------------------------------
    u_dp_latency_adapter_rx : ENTITY dp_lib.dp_latency_adapter
    GENERIC MAP (
      g_in_latency  => 1,
      g_out_latency => 0 
    )
    PORT MAP (
      clk       => kernel_clk,
      rst       => kernel_reset,

      snk_in    => dp_fifo_dc_src_out_arr(mac),
      snk_out   => dp_fifo_dc_src_in_arr(mac),

      src_out   => dp_latency_adapter_rx_src_out_arr(mac),
      src_in    => dp_latency_adapter_rx_src_in_arr(mac)
    );

    ----------------------------------------------------------------------------
    -- Data mapping 
    ----------------------------------------------------------------------------
    -- Reverse byte order
    gen_rx_bytes: FOR I IN 0 TO 31 GENERATE
      src_out_arr(mac).data(8*(32-I) -1  DOWNTO 8*(31-I)) <= dp_latency_adapter_rx_src_out_arr(mac).data(8*(I+1) -1 DOWNTO 8*I);
    END GENERATE;

    -- Assign control signals to correct data fields.
    src_out_arr(mac).data(256) <= dp_latency_adapter_rx_src_out_arr(mac).sop;
    src_out_arr(mac).data(257) <= dp_latency_adapter_rx_src_out_arr(mac).eop;
    src_out_arr(mac).data(263 DOWNTO 259) <= dp_latency_adapter_rx_src_out_arr(mac).empty(4 DOWNTO 0);
    src_out_arr(mac).valid <= dp_latency_adapter_rx_src_out_arr(mac).valid;
    dp_latency_adapter_rx_src_in_arr(mac).ready <= src_in_arr(mac).ready;
    dp_latency_adapter_rx_src_in_arr(mac).xon <= '1';

    -------------------------------------------------------------------------------
    -- Reset signal generation
    -------------------------------------------------------------------------------
  
    u_common_areset_txmac : ENTITY common_lib.common_areset
    GENERIC MAP (
      g_rst_level => '1',
      g_delay_len => 3
    )
    PORT MAP (
      in_rst    => kernel_reset,
      clk       => clk_txmac_arr(mac),
      out_rst   => rst_txmac_arr(mac)
    );


    -------------------------------------------------------------------------------
    -- PLL for clock generation, every mac needs its own, due to clock nework limitations
    -------------------------------------------------------------------------------
    u_arria10_40g_atx_pll : arria10_40g_atx_pll
    port map (
      pll_cal_busy  => OPEN,
      pll_locked    => pll_locked_arr(mac),
      pll_powerdown => config_reset,
      pll_refclk0   => clk_ref_r,
      tx_serial_clk => serial_clk_arr(mac)
    );
  
    gen_serial_clk_arr : FOR i IN 0 TO 3 GENERATE
      serial_clk_2arr(mac)(i) <= serial_clk_arr(mac);
    END GENERATE;

  END GENERATE;


END str;

