-------------------------------------------------------------------------------
--
-- Copyright (C) 2019
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
-- . Reinier van der Walle
-- Purpose:
-- . Provide 10G ethernet I/O interface (BSP) for OpenCL kernel on Arria10
-- Description:
-- . This core consists of:
--   . An Intel/Altera 10G Low Latency MAC instance
--   . SOP/EOP insertion (kernel channel only carries data and valid)
--   . Dual clock FIFO 
--     . Clock domain transition between kernel_clk and clk_txmac
--     . Buffers full Ethernet packet (10G MAC requires uninterrupted packet)
--   . Clock (PLL) / reset generation
-- . Details:
--   . This core was developed for use on the Uniboard2b.
--   .
--   . The data field of the ST-avalon interface is also used to provide
--   . SOP, EOP and empty meta-data. The implementation of this is shown below.
--   +-----------+---------+--------------------------------------------------------+
--   | Bit range | Name    | Description                                            |
--   +-----------+---------+--------------------------------------------------------+
--   | [0:15]    | payload | ADC channel 0 sample                                   |
--   +-----------+---------+--------------------------------------------------------+
LIBRARY IEEE, common_lib, dp_lib, tech_pll_lib, technology_lib, tech_jesd204b_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;

ENTITY ta2_unb2b_jesd204b IS       
  PORT (      
    config_clk       : IN  STD_LOGIC; -- 100MHz clk for reconfig block and status interface
    config_reset     : IN  STD_LOGIC;

    -- MM Control
    jesd204b_mosi : IN  t_mem_mosi;  
    jesd204b_miso : OUT t_mem_miso;

    -- JESD204B external signals
    jesd204b_refclk       : IN STD_LOGIC := '0';                             -- Reference clock. For AD9683 use 200MHz direct from clock reference pin
    jesd204b_sysref       : IN STD_LOGIC := '0';                             -- SYSREF should drive ADC and FPGA with correct phase wrt jesd204b_device_clk 
    jesd204b_sync_n_arr   : OUT STD_LOGIC_VECTOR(0 DOWNTO 0); -- output to control ADC initialization/syncronization phase
 
    serial_rx_arr         : IN  STD_LOGIC_VECTOR(0 DOWNTO 0);

    kernel_clk       : IN  STD_LOGIC; -- Kernel clock (runs the kernel_* I/O below)
    kernel_reset     : IN  STD_LOGIC;

    kernel_src_data  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0); -- RX Data to kernel
    kernel_src_valid : OUT STD_LOGIC; -- RX data valid signal to kernel
    kernel_src_ready : IN  STD_LOGIC -- Flow control from kernel

  );
END ta2_unb2b_jesd204b;


ARCHITECTURE str OF ta2_unb2b_jesd204b IS

  CONSTANT c_sim                           : BOOLEAN := FALSE;
  
  CONSTANT c_nof_connected_streams_jesd204b: NATURAL := 1;
  CONSTANT c_nof_streams_jesd204b          : NATURAL := 12;

  CONSTANT c_rx_fifo_size                  : NATURAL := 32; -- should be large enough

  SIGNAL dp_fifo_dc_rx_src_out             : t_dp_sosi; 
  SIGNAL dp_fifo_dc_rx_snk_in              : t_dp_sosi := c_dp_sosi_rst; 
  SIGNAL dp_fifo_dc_rx_src_in              : t_dp_siso;
  SIGNAL dp_fifo_dc_rx_snk_out             : t_dp_siso;

  SIGNAL dp_latency_adapter_rx_src_out     : t_dp_sosi; 
  SIGNAL dp_latency_adapter_rx_src_in      : t_dp_siso;

  SIGNAL jesd204b_rx_src_out_arr           : t_dp_sosi_arr(c_nof_streams_jesd204b-1 DOWNTO 0);
  SIGNAL jesd204b_frame_clk                : STD_LOGIC;
  SIGNAL jesd204b_rx_src_out_flat_w_sync   : t_dp_sosi;

  SIGNAL i_jesd204b_sync_n_arr             : STD_LOGIC_VECTOR(c_nof_streams_jesd204b-1 DOWNTO 0);
  SIGNAL jesd204b_serial_rx_arr            : STD_LOGIC_VECTOR(c_nof_streams_jesd204b-1 DOWNTO 0) := (OTHERS => '0');
BEGIN

  jesd204b_sync_n_arr <= i_jesd204b_sync_n_arr(c_nof_connected_streams_jesd204b -1 DOWNTO 0);
  jesd204b_serial_rx_arr(c_nof_connected_streams_jesd204b -1 DOWNTO 0) <= serial_rx_arr;

  u_jesd204b: ENTITY tech_jesd204b_lib.tech_jesd204b 
  GENERIC MAP(
    g_sim                => c_sim,                
    g_nof_channels       => c_nof_streams_jesd204b    
  )
  PORT MAP(
    jesd204b_refclk      => jesd204b_refclk,   
    jesd204b_sysref      => jesd204b_sysref,   
    jesd204b_sync_n_arr  => i_jesd204b_sync_n_arr,   
  
    rx_src_out_arr       => jesd204b_rx_src_out_arr,          
    jesd204b_frame_clk   => jesd204b_frame_clk,          
  
    -- MM
    mm_clk               => config_clk,           
    mm_rst               => config_reset,           
  
    jesd204b_mosi        => jesd204b_mosi,         
    jesd204b_miso        => jesd204b_miso,         
  
     -- Serial
    serial_tx_arr        => open,
    serial_rx_arr        => jesd204b_serial_rx_arr
  );
    

  ---------------------------------------------------------------------------------------
  -- RX FIFO: adc_clk -> kernel_clk
  ---------------------------------------------------------------------------------------

  dp_fifo_dc_rx_snk_in.data(13 DOWNTO 0) <= jesd204b_rx_src_out_arr(0).data(15 DOWNTO 2);
  dp_fifo_dc_rx_snk_in.data(14) <= jesd204b_rx_src_out_arr(0).data(15);
  dp_fifo_dc_rx_snk_in.data(15) <= jesd204b_rx_src_out_arr(0).data(15);
  dp_fifo_dc_rx_snk_in.valid <= dp_fifo_dc_rx_snk_out.ready AND jesd204b_rx_src_out_arr(0).valid;

  u_dp_fifo_dc_rx : ENTITY dp_lib.dp_fifo_dc
  GENERIC MAP (
    g_technology  => c_tech_arria10_e1sg,
    g_data_w      => 16,
    g_empty_w     => 1,
    g_use_empty   => FALSE,
    g_use_ctrl    => FALSE,
    g_fifo_size   => c_rx_fifo_size
  )
  PORT MAP (
    wr_rst      => kernel_reset,
    wr_clk      => jesd204b_frame_clk,
    rd_rst      => kernel_reset,
    rd_clk      => kernel_clk,
  
    snk_out     => dp_fifo_dc_rx_snk_out,
    snk_in      => dp_fifo_dc_rx_snk_in,
  
    src_in      => dp_fifo_dc_rx_src_in, 
    src_out     => dp_fifo_dc_rx_src_out
  );   


  ----------------------------------------------------------------------------
  -- Latency adapter: adapt RL=1 (Upstream) to RL=0 (OpenCL kernel).
  ----------------------------------------------------------------------------
  u_dp_latency_adapter_rx : ENTITY dp_lib.dp_latency_adapter
  GENERIC MAP (
    g_in_latency  => 1,
    g_out_latency => 0 
  )
  PORT MAP (
    clk       => kernel_clk,
    rst       => kernel_reset,

    snk_in    => dp_fifo_dc_rx_src_out, 
    snk_out   => dp_fifo_dc_rx_src_in, 

    src_out   => dp_latency_adapter_rx_src_out, 
    src_in    => dp_latency_adapter_rx_src_in 
  );
      

  ----------------------------------------------------------------------------
  -- Data mapping 
  ----------------------------------------------------------------------------
  -- Reverse byte order
  --gen_rx_bytes: FOR I IN 0 TO c_halfword_sz-1 GENERATE
  --  kernel_src_data(c_byte_w*(c_halfword_sz-I) -1  DOWNTO c_byte_w*(c_halfword_sz-1-I)) <= dp_latency_adapter_rx_src_out.data(c_byte_w*(I+1) -1 DOWNTO c_byte_w*I);
  --END GENERATE;

  kernel_src_data <= dp_latency_adapter_rx_src_out.data(15 DOWNTO 0);

  kernel_src_valid <= dp_latency_adapter_rx_src_out.valid;
  dp_latency_adapter_rx_src_in.ready <= kernel_src_ready;
  dp_latency_adapter_rx_src_in.xon <= '1';


END str;

