-------------------------------------------------------------------------------
--
-- Copyright (C) 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb2b_board_lib, technology_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE unb2b_board_lib.unb2b_board_pkg.ALL;
USE unb2b_board_lib.unb2b_board_peripherals_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.top_components_pkg.ALL;

ENTITY top IS
  GENERIC (
    g_design_name       : STRING  := "ta2_unb2b_bsp";
    g_design_note       : STRING  := "UNUSED";
    g_technology        : NATURAL := c_tech_arria10_e1sg;
    g_sim               : BOOLEAN := FALSE; --Overridden by TB
    g_sim_unb_nr        : NATURAL := 0;
    g_sim_node_nr       : NATURAL := 0;
    g_stamp_date        : NATURAL := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time        : NATURAL := 0;  -- Time (HHMMSS)   -- set by QSF
    g_revision_id       : STRING  := "";  -- revision_id, commit hash (first 9 chars) or number
    g_factory_image     : BOOLEAN := FALSE;
    g_protect_addr_range: BOOLEAN := FALSE
  );
  PORT (
    -- GENERAL
    CLK          : IN    STD_LOGIC; -- System Clock
    PPS          : IN    STD_LOGIC; -- System Sync
    WDI          : OUT   STD_LOGIC; -- Watchdog Clear
    INTA         : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB         : INOUT STD_LOGIC; -- FPGA interconnect line

    -- Others
    VERSION      : IN    STD_LOGIC_VECTOR(c_unb2b_board_aux.version_w-1 DOWNTO 0);
    ID           : IN    STD_LOGIC_VECTOR(c_unb2b_board_aux.id_w-1 DOWNTO 0);
    TESTIO       : INOUT STD_LOGIC_VECTOR(c_unb2b_board_aux.testio_w-1 DOWNTO 0);
    
    -- I2C Interface to Sensors
    SENS_SC      : INOUT STD_LOGIC;
    SENS_SD      : INOUT STD_LOGIC;

    PMBUS_SC     : INOUT STD_LOGIC;
    PMBUS_SD     : INOUT STD_LOGIC;
    PMBUS_ALERT  : IN    STD_LOGIC := '0';
  
    -- 1GbE Control Interface
    ETH_CLK      : IN    STD_LOGIC;
    ETH_SGIN     : IN    STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 DOWNTO 0);
    ETH_SGOUT    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 DOWNTO 0);

    -- Transceiver clocks
    SA_CLK       : IN    STD_LOGIC := '0'; -- Clock 10GbE front (qsfp) and ring lines

    -- front transceivers
    QSFP_0_RX    : IN    STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp.bus_w-1 DOWNTO 0) := (OTHERS=>'0');
    QSFP_0_TX    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp.bus_w-1 DOWNTO 0);
    QSFP_1_RX    : IN    STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp.bus_w-1 DOWNTO 0) := (OTHERS=>'0');
    QSFP_1_TX    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp.bus_w-1 DOWNTO 0);

    -- ring transceivers
    RING_0_RX    : IN    STD_LOGIC_VECTOR(c_unb2b_board_tr_ring.bus_w-1 DOWNTO 0) := (OTHERS => '0');
    RING_0_TX    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_ring.bus_w-1 DOWNTO 0);
    RING_1_RX    : IN    STD_LOGIC_VECTOR(c_unb2b_board_tr_ring.bus_w-1 DOWNTO 0) := (OTHERS=>'0');
    RING_1_TX    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_ring.bus_w-1 DOWNTO 0);

     -- back transceivers
    BCK_RX       : IN    STD_LOGIC_VECTOR(0 DOWNTO 0);
    BCK_REF_CLK  : IN    STD_LOGIC; -- Use as JESD204B_REFCLK
 
    -- jesd204b syncronization signals
    JESD204B_SYSREF : IN    STD_LOGIC;
    JESD204B_SYNC   : OUT   STD_LOGIC_VECTOR(0 DOWNTO 0);

    -- LEDs
    QSFP_LED     : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp_nof_leds-1 DOWNTO 0)
  );
END top;


ARCHITECTURE str OF top IS

  -- Firmware version x.y
  CONSTANT c_fw_version             : t_unb2b_board_fw_version := (1, 1);
  CONSTANT c_mm_clk_freq            : NATURAL := c_unb2b_board_mm_clk_freq_100M;
  
  -- 1 GbE
  CONSTANT c_use_1GbE_udp_offload   : BOOLEAN := TRUE;
  CONSTANT c_nof_streams_1GbE       : NATURAL := 1;

  -- 10GbE
  CONSTANT c_nof_qsfp_bus           : NATURAL := 2;
  CONSTANT c_nof_ring_bus           : NATURAL := 2;
  CONSTANT c_ring_bus_w             : NATURAL := c_unb2b_board_tr_ring.bus_w;
  CONSTANT c_nof_streams_qsfp       : NATURAL := c_unb2b_board_tr_qsfp.bus_w*c_nof_qsfp_bus;
  CONSTANT c_nof_streams_ring       : NATURAL := c_unb2b_board_tr_ring.bus_w*c_nof_ring_bus;

  -- 40GbE
  CONSTANT c_nof_40GbE_IP           : NATURAL := 3;

  -- System
  SIGNAL cs_sim                     : STD_LOGIC;
  SIGNAL xo_ethclk                  : STD_LOGIC;
  SIGNAL xo_rst                     : STD_LOGIC;
  SIGNAL xo_rst_n                   : STD_LOGIC;
  SIGNAL mm_clk                     : STD_LOGIC;
  SIGNAL mm_rst                     : STD_LOGIC;
  
  SIGNAL st_rst                     : STD_LOGIC;
  SIGNAL st_clk                     : STD_LOGIC;

  -- PIOs
  SIGNAL pout_wdi                   : STD_LOGIC;

  -- WDI override
  SIGNAL reg_wdi_mosi               : t_mem_mosi;
  SIGNAL reg_wdi_miso               : t_mem_miso;

  -- PPSH
  SIGNAL reg_ppsh_mosi              : t_mem_mosi;
  SIGNAL reg_ppsh_miso              : t_mem_miso;
  
  -- UniBoard system info
  SIGNAL reg_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso   : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso   : t_mem_miso;

  -- UniBoard I2C sens
  SIGNAL reg_unb_sens_mosi          : t_mem_mosi;
  SIGNAL reg_unb_sens_miso          : t_mem_miso;

  -- pm bus
  SIGNAL reg_unb_pmbus_mosi         : t_mem_mosi;
  SIGNAL reg_unb_pmbus_miso         : t_mem_miso;

  -- FPGA sensors
  SIGNAL reg_fpga_temp_sens_mosi    : t_mem_mosi;
  SIGNAL reg_fpga_temp_sens_miso    : t_mem_miso;
  SIGNAL reg_fpga_voltage_sens_mosi : t_mem_mosi;
  SIGNAL reg_fpga_voltage_sens_miso : t_mem_miso;

  -- eth1g
  SIGNAL eth1g_mm_rst               : STD_LOGIC;
  SIGNAL eth1g_tse_mosi             : t_mem_mosi;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso             : t_mem_miso;
  SIGNAL eth1g_reg_mosi             : t_mem_mosi;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso             : t_mem_miso;
  SIGNAL eth1g_reg_interrupt        : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi             : t_mem_mosi;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso             : t_mem_miso;

  -- Interface: 1GbE UDP streaming ports
  SIGNAL eth1g_udp_tx_sosi_arr      : t_dp_sosi_arr(c_nof_streams_1GbE-1 DOWNTO 0);
  SIGNAL eth1g_udp_tx_siso_arr      : t_dp_siso_arr(c_nof_streams_1GbE-1 DOWNTO 0);
  SIGNAL eth1g_udp_rx_sosi_arr      : t_dp_sosi_arr(c_nof_streams_1GbE-1 DOWNTO 0);
  SIGNAL eth1g_udp_rx_siso_arr      : t_dp_siso_arr(c_nof_streams_1GbE-1 DOWNTO 0);

  -- EPCS read
  SIGNAL reg_dpmm_data_mosi         : t_mem_mosi;
  SIGNAL reg_dpmm_data_miso         : t_mem_miso;
  SIGNAL reg_dpmm_ctrl_mosi         : t_mem_mosi;
  SIGNAL reg_dpmm_ctrl_miso         : t_mem_miso;

  -- EPCS write
  SIGNAL reg_mmdp_data_mosi         : t_mem_mosi;
  SIGNAL reg_mmdp_data_miso         : t_mem_miso;
  SIGNAL reg_mmdp_ctrl_mosi         : t_mem_mosi;
  SIGNAL reg_mmdp_ctrl_miso         : t_mem_miso;

  -- EPCS status/control
  SIGNAL reg_epcs_mosi              : t_mem_mosi;
  SIGNAL reg_epcs_miso              : t_mem_miso;

  -- Remote Update
  SIGNAL reg_remu_mosi              : t_mem_mosi;
  SIGNAL reg_remu_miso              : t_mem_miso;

  -- JESD204b
  SIGNAL reg_ta2_unb2b_jesd204b_mosi : t_mem_mosi;
  SIGNAL reg_ta2_unb2b_jesd204b_miso : t_mem_miso;

  -- 10GbE
  SIGNAL i_QSFP_TX                  : t_unb2b_board_qsfp_bus_2arr(c_nof_qsfp_bus-1 DOWNTO 0); 
  SIGNAL i_QSFP_RX                  : t_unb2b_board_qsfp_bus_2arr(c_nof_qsfp_bus-1 DOWNTO 0);

  SIGNAL i_RING_TX                  : t_unb2b_board_ring_bus_2arr(c_nof_ring_bus-1 DOWNTO 0); 
  SIGNAL i_RING_RX                  : t_unb2b_board_ring_bus_2arr(c_nof_ring_bus-1 DOWNTO 0);
  
  SIGNAL unb2b_board_front_io_serial_tx_arr : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL unb2b_board_front_io_serial_rx_arr : STD_LOGIC_VECTOR(c_nof_streams_qsfp-1 DOWNTO 0);
 
  SIGNAL unb2b_board_ring_io_serial_tx_arr : STD_LOGIC_VECTOR(c_nof_streams_ring-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL unb2b_board_ring_io_serial_rx_arr : STD_LOGIC_VECTOR(c_nof_streams_ring-1 DOWNTO 0);

  -- QSFP leds
  SIGNAL qsfp_green_led_arr         : STD_LOGIC_VECTOR(c_nof_qsfp_bus+c_nof_ring_bus-1 DOWNTO 0);
  SIGNAL qsfp_red_led_arr           : STD_LOGIC_VECTOR(c_nof_qsfp_bus+c_nof_ring_bus-1 DOWNTO 0);
  --SIGNAL unb2b_board_qsfp_leds_tx_src_in_arr : t_dp_siso_arr(c_nof_qsfp_bus*c_quad-1 DOWNTO 0) := (OTHERS => c_dp_siso_rst);
  SIGNAL unb2b_board_qsfp_leds_tx_src_in_arr : t_dp_siso_arr((c_nof_qsfp_bus+c_nof_qsfp_bus)*c_quad-1 DOWNTO 0) := (OTHERS => c_dp_siso_rst);

  SIGNAL i_reset_n         : STD_LOGIC;
  SIGNAL i_kernel_rst    : STD_LOGIC;

  -- OpenCL kernel
  SIGNAL board_kernel_clk_clk                         : std_logic;                    
  SIGNAL board_kernel_clk2x_clk                       : std_logic;                   
  SIGNAL board_kernel_reset_reset_n                   : std_logic;                   
  SIGNAL board_kernel_reset_reset_n_in                : std_logic;                   
 
  SIGNAL board_kernel_cra_waitrequest                 : std_logic;                     
  SIGNAL board_kernel_cra_readdata                    : std_logic_vector(63 downto 0); 
  SIGNAL board_kernel_cra_readdatavalid               : std_logic;                     
  SIGNAL board_kernel_cra_burstcount                  : std_logic_vector(0 downto 0); 
  SIGNAL board_kernel_cra_writedata                   : std_logic_vector(63 downto 0);
  SIGNAL board_kernel_cra_address                     : std_logic_vector(29 downto 0);
  SIGNAL board_kernel_cra_write                       : std_logic;                    
  SIGNAL board_kernel_cra_read                        : std_logic;                    
  SIGNAL board_kernel_cra_byteenable                  : std_logic_vector(7 downto 0); 
  SIGNAL board_kernel_cra_debugaccess                 : std_logic;                    

  SIGNAL board_kernel_irq_irq                         : std_logic_vector(0 downto 0); 

  SIGNAL board_kernel_register_mem_address            : std_logic_vector(6 downto 0); --  := (others => 'X'); -- address
  SIGNAL board_kernel_register_mem_clken              : std_logic; --                     := 'X';             -- clken
  SIGNAL board_kernel_register_mem_chipselect         : std_logic; --                     := 'X';             -- chipselect
  SIGNAL board_kernel_register_mem_write              : std_logic; --                     := 'X';             -- write
  SIGNAL board_kernel_register_mem_readdata           : std_logic_vector(255 downto 0);                    -- readdata
  SIGNAL board_kernel_register_mem_writedata          : std_logic_vector(255 downto 0); -- := (others => 'X'); -- writedata
  SIGNAL board_kernel_register_mem_byteenable         : std_logic_vector(31 downto 0); --  := (others => 'X'); -- byteenable

  SIGNAL ta2_unb2b_40GbE_src_out_arr                  : t_dp_sosi_arr(c_nof_40GbE_IP-1 DOWNTO 0);
  SIGNAL ta2_unb2b_40GbE_src_in_arr                   : t_dp_siso_arr(c_nof_40GbE_IP-1 DOWNTO 0);
  SIGNAL ta2_unb2b_40GbE_snk_out_arr                  : t_dp_siso_arr(c_nof_40GbE_IP-1 DOWNTO 0);
  SIGNAL ta2_unb2b_40GbE_snk_in_arr                   : t_dp_sosi_arr(c_nof_40GbE_IP-1 DOWNTO 0);
  SIGNAL ta2_unb2b_40GbE_tx_serial_r                  : STD_LOGIC_VECTOR(4*c_nof_40GbE_IP -1 DOWNTO 0);
  SIGNAL ta2_unb2b_40GbE_rx_serial_r                  : STD_LOGIC_VECTOR(4*c_nof_40GbE_IP -1 DOWNTO 0);

  SIGNAL board_kernel_stream_src_10GbE_data           : std_logic_vector(71 downto 0); 
  SIGNAL board_kernel_stream_src_10GbE_valid          : std_logic; 
  SIGNAL board_kernel_stream_src_10GbE_ready          : std_logic; 
  SIGNAL board_kernel_stream_snk_10GbE_data           : std_logic_vector(71 downto 0); 
  SIGNAL board_kernel_stream_snk_10GbE_valid          : std_logic; 
  SIGNAL board_kernel_stream_snk_10GbE_ready          : std_logic;
  SIGNAL ta2_unb2b_10gbe_rx_status_rx_status          : std_logic;                                         -- rx_status

  SIGNAL board_kernel_stream_src_1GbE_data           : std_logic_vector(39 downto 0); 
  SIGNAL board_kernel_stream_src_1GbE_valid          : std_logic; 
  SIGNAL board_kernel_stream_src_1GbE_ready          : std_logic; 
  SIGNAL board_kernel_stream_snk_1GbE_data           : std_logic_vector(39 downto 0); 
  SIGNAL board_kernel_stream_snk_1GbE_valid          : std_logic; 
  SIGNAL board_kernel_stream_snk_1GbE_ready          : std_logic;

  SIGNAL board_kernel_stream_src_ADC_data           : std_logic_vector(15 downto 0); 
  SIGNAL board_kernel_stream_src_ADC_valid          : std_logic; 
  SIGNAL board_kernel_stream_src_ADC_ready          : std_logic; 

BEGIN
  ------------
  -- Front IO
  ------------

  -- put the QSFP_TX/RX ports into arrays 
  i_QSFP_RX(0) <= QSFP_0_RX;
  i_QSFP_RX(1) <= QSFP_1_RX;

  QSFP_0_TX <= i_QSFP_TX(0);
  QSFP_1_TX <= i_QSFP_TX(1);

  u_unb2b_board_front_io : ENTITY unb2b_board_lib.unb2b_board_front_io
  GENERIC MAP (
    g_nof_qsfp_bus => c_nof_qsfp_bus
  )
  PORT MAP (
    serial_tx_arr => unb2b_board_front_io_serial_tx_arr,
    serial_rx_arr => unb2b_board_front_io_serial_rx_arr,

    --green_led_arr => qsfp_green_led_arr(c_nof_qsfp_bus-1 DOWNTO 0),
    --red_led_arr   => qsfp_red_led_arr(c_nof_qsfp_bus-1 DOWNTO 0),

    QSFP_RX    => i_QSFP_RX,
    QSFP_TX    => i_QSFP_TX --,

    --QSFP_LED   => QSFP_LED
  );

  ------------------------
  -- qsfp LEDs controller 
  ------------------------
  unb2b_board_qsfp_leds_tx_src_in_arr(0).xon <= ta2_unb2b_10gbe_rx_status_rx_status;
  unb2b_board_qsfp_leds_tx_src_in_arr(4).xon <= ta2_unb2b_40GbE_snk_out_arr(0).xon;
  -- ring LED indicator
  unb2b_board_qsfp_leds_tx_src_in_arr(8).xon <= ta2_unb2b_40GbE_snk_out_arr(1).xon;
  unb2b_board_qsfp_leds_tx_src_in_arr(12).xon <= ta2_unb2b_40GbE_snk_out_arr(2).xon;

  u_unb2b_board_qsfp_leds : ENTITY unb2b_board_lib.unb2b_board_qsfp_leds
  GENERIC MAP (
    g_sim             => g_sim,
    g_factory_image   => g_factory_image,
    g_nof_qsfp        => c_nof_qsfp_bus+c_nof_ring_bus,
    g_pulse_us        => 1000 / (10**9 / c_mm_clk_freq) -- nof clk cycles to get us period
  )
  PORT MAP (
    rst               => mm_rst,
    clk               => mm_clk,

    tx_siso_arr       => unb2b_board_qsfp_leds_tx_src_in_arr,

    green_led_arr     => qsfp_green_led_arr(c_nof_qsfp_bus+c_nof_ring_bus-1 DOWNTO 0),
    red_led_arr       => qsfp_red_led_arr(c_nof_qsfp_bus+c_nof_ring_bus-1 DOWNTO 0)
  );

  gen_leds : FOR i IN 0 TO c_nof_qsfp_bus+c_nof_ring_bus-1 GENERATE
    QSFP_LED(i*2)   <=  qsfp_green_led_arr(i);
    QSFP_LED(i*2+1) <=  qsfp_red_led_arr(i);
  END GENERATE;



  ------------
  -- RING IO
  ------------
  i_RING_RX(0) <= RING_0_RX;
  i_RING_RX(1) <= RING_1_RX;
  RING_0_TX <= i_RING_TX(0);
  RING_1_TX <= i_RING_TX(1);

  u_ring_io : ENTITY unb2b_board_lib.unb2b_board_ring_io
  GENERIC MAP (
    g_nof_ring_bus => c_nof_ring_bus
  )
  PORT MAP (
    serial_tx_arr => unb2b_board_ring_io_serial_tx_arr,
    serial_rx_arr => unb2b_board_ring_io_serial_rx_arr,
    RING_RX => i_RING_RX,
    RING_TX => i_RING_TX
  );

  ---------
  -- 40GbE 
  ---------
  -- Front QSFP 1 40GbE Interface, IP index = 0
  unb2b_board_front_io_serial_tx_arr(7 DOWNTO 4) <= ta2_unb2b_40GbE_tx_serial_r(3 DOWNTO 0);
  ta2_unb2b_40GbE_rx_serial_r(3 DOWNTO 0) <= unb2b_board_front_io_serial_rx_arr(7 DOWNTO 4);
  
  -- Ring 0 (to the left), IP index = 1
  unb2b_board_ring_io_serial_tx_arr(3 DOWNTO 0) <= ta2_unb2b_40GbE_tx_serial_r(7 DOWNTO 4);
  ta2_unb2b_40GbE_rx_serial_r(7 DOWNTO 4) <= unb2b_board_ring_io_serial_rx_arr(3 DOWNTO 0);

  -- Ring 1 (to the right), IP index = 2
  unb2b_board_ring_io_serial_tx_arr(3+c_ring_bus_w DOWNTO c_ring_bus_w) <= ta2_unb2b_40GbE_tx_serial_r(11 DOWNTO 8);
  ta2_unb2b_40GbE_rx_serial_r(11 DOWNTO 8) <= unb2b_board_ring_io_serial_rx_arr(3+c_ring_bus_w DOWNTO c_ring_bus_w);

  u_ta2_unb2b_40GbE : ENTITY work.ta2_unb2b_40GbE
  GENERIC MAP (
    g_nof_mac => c_nof_40GbE_IP
  )
  PORT MAP (
    config_clk       => mm_clk, 
    config_reset     => mm_rst, 

    clk_ref_r        => SA_CLK, 

    tx_serial_r      => ta2_unb2b_40GbE_tx_serial_r, 
    rx_serial_r      => ta2_unb2b_40GbE_rx_serial_r, 

    kernel_clk       => board_kernel_clk_clk, 
    kernel_reset     => i_kernel_rst,

    src_out_arr      => ta2_unb2b_40GbE_src_out_arr, 
    src_in_arr       => ta2_unb2b_40GbE_src_in_arr, 
    snk_out_arr      => ta2_unb2b_40GbE_snk_out_arr, 
    snk_in_arr       => ta2_unb2b_40GbE_snk_in_arr
  );

  ----------
  -- 10GbE
  ----------
  u_ta2_unb2b_10GbE : ENTITY work.ta2_unb2b_10GbE
  PORT MAP (
    config_reset     => mm_rst, 

    clk_ref_r        => SA_CLK, 

    tx_serial_r      => unb2b_board_front_io_serial_tx_arr(0),
    rx_serial_r      => unb2b_board_front_io_serial_rx_arr(0), 

    kernel_clk       => board_kernel_clk_clk, 
    kernel_reset     => i_kernel_rst, 

    kernel_src_data  => board_kernel_stream_src_10GbE_data, 
    kernel_src_valid => board_kernel_stream_src_10GbE_valid, 
    kernel_src_ready => board_kernel_stream_src_10GbE_ready, 

    kernel_snk_data  => board_kernel_stream_snk_10GbE_data, 
    kernel_snk_valid => board_kernel_stream_snk_10GbE_valid, 
    kernel_snk_ready => board_kernel_stream_snk_10GbE_ready, 

    rx_status        => ta2_unb2b_10gbe_rx_status_rx_status 
  );


  -----------------------------
  -- 1GbE Monitoring & Control
  -----------------------------
  u_ta2_unb2b_1GbE_mc : ENTITY work.ta2_unb2b_1GbE_mc
  PORT MAP (
    st_clk           => st_clk,
    st_rst           => st_rst,

    udp_tx_sosi      => eth1g_udp_tx_sosi_arr(0),
    udp_tx_siso      => eth1g_udp_tx_siso_arr(0),
    udp_rx_sosi      => eth1g_udp_rx_sosi_arr(0),
    udp_rx_siso      => eth1g_udp_rx_siso_arr(0),

    kernel_clk       => board_kernel_clk_clk, 
    kernel_reset     => i_kernel_rst, 

    kernel_src_data  => board_kernel_stream_src_1GbE_data, 
    kernel_src_valid => board_kernel_stream_src_1GbE_valid, 
    kernel_src_ready => board_kernel_stream_src_1GbE_ready, 
    kernel_snk_data  => board_kernel_stream_snk_1GbE_data, 
    kernel_snk_valid => board_kernel_stream_snk_1GbE_valid, 
    kernel_snk_ready => board_kernel_stream_snk_1GbE_ready

  );

  ----------
  -- ADC
  ----------
  u_ta2_unb2b_jesd204b : ENTITY work.ta2_unb2b_jesd204b
  PORT MAP(      
    config_clk    => mm_clk,    
    config_reset  => mm_rst,
   
    jesd204b_mosi => reg_ta2_unb2b_jesd204b_mosi,
    jesd204b_miso => reg_ta2_unb2b_jesd204b_miso,

    -- JESD204B external signals
    jesd204b_refclk       => BCK_REF_CLK, 
    jesd204b_sysref       => JESD204B_SYSREF,  
    jesd204b_sync_n_arr   => JESD204B_SYNC,
 
    serial_rx_arr         => BCK_RX, 

    kernel_clk            => board_kernel_clk_clk, 
    kernel_reset          => i_kernel_rst, 

    kernel_src_data       => board_kernel_stream_src_ADC_data,
    kernel_src_valid      => board_kernel_stream_src_ADC_valid,
    kernel_src_ready      => board_kernel_stream_src_ADC_ready 

  );


  -----------------------------------------------------------------------------
  -- Freeze wrapper instantiation 
  -----------------------------------------------------------------------------
  freeze_wrapper_inst : freeze_wrapper
  PORT MAP(
    board_kernel_clk_clk                 => board_kernel_clk_clk,  
    board_kernel_clk2x_clk               => board_kernel_clk2x_clk,
    board_kernel_reset_reset_n           => board_kernel_reset_reset_n_in,
    board_kernel_irq_irq                 => board_kernel_irq_irq,
    board_kernel_cra_waitrequest         => board_kernel_cra_waitrequest,
    board_kernel_cra_readdata            => board_kernel_cra_readdata,
    board_kernel_cra_readdatavalid       => board_kernel_cra_readdatavalid,
    board_kernel_cra_burstcount          => board_kernel_cra_burstcount,
    board_kernel_cra_writedata           => board_kernel_cra_writedata,
    board_kernel_cra_address             => board_kernel_cra_address,
    board_kernel_cra_write               => board_kernel_cra_write,
    board_kernel_cra_read                => board_kernel_cra_read,
    board_kernel_cra_byteenable          => board_kernel_cra_byteenable,
    board_kernel_cra_debugaccess         => board_kernel_cra_debugaccess,
    board_kernel_register_mem_address    => board_kernel_register_mem_address,
    board_kernel_register_mem_clken      => board_kernel_register_mem_clken, 
    board_kernel_register_mem_chipselect => board_kernel_register_mem_chipselect,
    board_kernel_register_mem_write      => board_kernel_register_mem_write, 
    board_kernel_register_mem_readdata   => board_kernel_register_mem_readdata,
    board_kernel_register_mem_writedata  => board_kernel_register_mem_writedata,
    board_kernel_register_mem_byteenable => board_kernel_register_mem_byteenable,  

    board_kernel_stream_src_40GbE_data   => ta2_unb2b_40GbE_src_out_arr(0).data(263 DOWNTO 0),
    board_kernel_stream_src_40GbE_valid  => ta2_unb2b_40GbE_src_out_arr(0).valid,
    board_kernel_stream_src_40GbE_ready  => ta2_unb2b_40GbE_src_in_arr(0).ready,
    board_kernel_stream_snk_40GbE_data   => ta2_unb2b_40GbE_snk_in_arr(0).data(263 DOWNTO 0),
    board_kernel_stream_snk_40GbE_valid  => ta2_unb2b_40GbE_snk_in_arr(0).valid,
    board_kernel_stream_snk_40GbE_ready  => ta2_unb2b_40GbE_snk_out_arr(0).ready,

    board_kernel_stream_src_40GbE_ring_0_data   => ta2_unb2b_40GbE_src_out_arr(1).data(263 DOWNTO 0),
    board_kernel_stream_src_40GbE_ring_0_valid  => ta2_unb2b_40GbE_src_out_arr(1).valid,
    board_kernel_stream_src_40GbE_ring_0_ready  => ta2_unb2b_40GbE_src_in_arr(1).ready,
    board_kernel_stream_snk_40GbE_ring_0_data   => ta2_unb2b_40GbE_snk_in_arr(1).data(263 DOWNTO 0),
    board_kernel_stream_snk_40GbE_ring_0_valid  => ta2_unb2b_40GbE_snk_in_arr(1).valid,
    board_kernel_stream_snk_40GbE_ring_0_ready  => ta2_unb2b_40GbE_snk_out_arr(1).ready,

    board_kernel_stream_src_40GbE_ring_1_data   => ta2_unb2b_40GbE_src_out_arr(2).data(263 DOWNTO 0),
    board_kernel_stream_src_40GbE_ring_1_valid  => ta2_unb2b_40GbE_src_out_arr(2).valid,
    board_kernel_stream_src_40GbE_ring_1_ready  => ta2_unb2b_40GbE_src_in_arr(2).ready,
    board_kernel_stream_snk_40GbE_ring_1_data   => ta2_unb2b_40GbE_snk_in_arr(2).data(263 DOWNTO 0),
    board_kernel_stream_snk_40GbE_ring_1_valid  => ta2_unb2b_40GbE_snk_in_arr(2).valid,
    board_kernel_stream_snk_40GbE_ring_1_ready  => ta2_unb2b_40GbE_snk_out_arr(2).ready,

    board_kernel_stream_src_10GbE_data   => board_kernel_stream_src_10GbE_data,
    board_kernel_stream_src_10GbE_valid  => board_kernel_stream_src_10GbE_valid,
    board_kernel_stream_src_10GbE_ready  => board_kernel_stream_src_10GbE_ready,
    board_kernel_stream_snk_10GbE_data   => board_kernel_stream_snk_10GbE_data,
    board_kernel_stream_snk_10GbE_valid  => board_kernel_stream_snk_10GbE_valid,
    board_kernel_stream_snk_10GbE_ready  => board_kernel_stream_snk_10GbE_ready,

    board_kernel_stream_src_1GbE_data    => board_kernel_stream_src_1GbE_data,
    board_kernel_stream_src_1GbE_valid   => board_kernel_stream_src_1GbE_valid,
    board_kernel_stream_src_1GbE_ready   => board_kernel_stream_src_1GbE_ready,
    board_kernel_stream_snk_1GbE_data    => board_kernel_stream_snk_1GbE_data,
    board_kernel_stream_snk_1GbE_valid   => board_kernel_stream_snk_1GbE_valid,
    board_kernel_stream_snk_1GbE_ready   => board_kernel_stream_snk_1GbE_ready,

    board_kernel_stream_src_ADC_data     => board_kernel_stream_src_ADC_data,
    board_kernel_stream_src_ADC_valid    => board_kernel_stream_src_ADC_valid,
    board_kernel_stream_src_ADC_ready    => board_kernel_stream_src_ADC_ready

  );

  i_reset_n <= NOT mm_rst;
  i_kernel_rst <= NOT board_kernel_reset_reset_n;

  -- Kernel should start later than BSP
  u_common_areset : ENTITY common_lib.common_areset
  GENERIC MAP (
    g_rst_level => '0',
    g_delay_len => 9
  )
  PORT MAP (
    in_rst  => i_kernel_rst,
    clk     => board_kernel_clk_clk,
    out_rst => board_kernel_reset_reset_n_in
  );


  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl_unb2b_board : ENTITY unb2b_board_lib.ctrl_unb2b_board
  GENERIC MAP (
    g_sim                     => g_sim,
    g_technology              => g_technology,
    g_design_name             => g_design_name,
    g_design_note             => g_design_note,
    g_stamp_date              => g_stamp_date,
    g_stamp_time              => g_stamp_time,
    g_revision_id             => g_revision_id,
    g_fw_version              => c_fw_version,
    g_mm_clk_freq             => c_mm_clk_freq,
    g_eth_clk_freq            => c_unb2b_board_eth_clk_freq_125M,
    g_udp_offload             => c_use_1GbE_udp_offload,
    g_udp_offload_nof_streams => c_nof_streams_1GbE,
    g_aux                     => c_unb2b_board_aux,
    g_factory_image           => g_factory_image,
    g_protect_addr_range      => g_protect_addr_range
  )
  PORT MAP (
    -- Clock an reset signals
    cs_sim                   => cs_sim,
    xo_ethclk                => xo_ethclk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => mm_clk,
    mm_rst                   => mm_rst,

    dp_rst                   => st_rst,
    dp_clk                   => st_clk,
    dp_pps                   => OPEN,
    dp_rst_in                => st_rst,
    dp_clk_in                => st_clk,
    
    -- Toggle WDI
    pout_wdi                 => pout_wdi,

    -- MM buses
    -- REMU
    reg_remu_mosi            => reg_remu_mosi,
    reg_remu_miso            => reg_remu_miso,

    -- EPCS read
    reg_dpmm_data_mosi       => reg_dpmm_data_mosi,
    reg_dpmm_data_miso       => reg_dpmm_data_miso,
    reg_dpmm_ctrl_mosi       => reg_dpmm_ctrl_mosi,
    reg_dpmm_ctrl_miso       => reg_dpmm_ctrl_miso,

    -- EPCS write
    reg_mmdp_data_mosi       => reg_mmdp_data_mosi,
    reg_mmdp_data_miso       => reg_mmdp_data_miso,
    reg_mmdp_ctrl_mosi       => reg_mmdp_ctrl_mosi,
    reg_mmdp_ctrl_miso       => reg_mmdp_ctrl_miso,

    -- EPCS status/control
    reg_epcs_mosi            => reg_epcs_mosi,
    reg_epcs_miso            => reg_epcs_miso,

    -- . Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,
    
    -- . System_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
    
    -- . UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,    
    
    -- . FPGA sensors
    reg_fpga_temp_sens_mosi  => reg_fpga_temp_sens_mosi,
    reg_fpga_temp_sens_miso  => reg_fpga_temp_sens_miso,
    reg_fpga_voltage_sens_mosi  => reg_fpga_voltage_sens_mosi,
    reg_fpga_voltage_sens_miso  => reg_fpga_voltage_sens_miso,

    reg_unb_pmbus_mosi       => reg_unb_pmbus_mosi,
    reg_unb_pmbus_miso       => reg_unb_pmbus_miso,

    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,
    
    -- eth1g
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,

    -- eth1g UDP streaming ports
    udp_tx_sosi_arr          =>  eth1g_udp_tx_sosi_arr,
    udp_tx_siso_arr          =>  eth1g_udp_tx_siso_arr,
    udp_rx_sosi_arr          =>  eth1g_udp_rx_sosi_arr,
    udp_rx_siso_arr          =>  eth1g_udp_rx_siso_arr,
        
    -- FPGA pins
    -- . General
    CLK                      => CLK,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    -- . Others
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    -- . I2C Interface to Sensors
    SENS_SC                  => SENS_SC,
    SENS_SD                  => SENS_SD,
    -- PM bus
    PMBUS_SC                 => PMBUS_SC,
    PMBUS_SD                 => PMBUS_SD,
    PMBUS_ALERT              => PMBUS_ALERT,

    -- . 1GbE Control Interface
    ETH_clk                  => ETH_CLK,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );


  -----------------------------------------------------------------------------
  -- Board qsys 
  -----------------------------------------------------------------------------
  board_inst : board
  PORT MAP (
      clk_clk                                   => mm_clk,
      reset_reset_n                             => i_reset_n,

      kernel_clk_clk                            => board_kernel_clk_clk,
      kernel_clk2x_clk                          => board_kernel_clk2x_clk,
      kernel_reset_reset_n                      => board_kernel_reset_reset_n,

      kernel_interface_sw_reset_in_reset        => mm_rst,  

      -- the_pio_wdi: toggled by NIOS II application unb_osy. Connects to WDI via ctrl_unb2b_board.
      pio_wdi_external_connection_export        => pout_wdi,

      avs_eth_0_reset_export                    => eth1g_mm_rst,
      avs_eth_0_tse_address_export              => eth1g_tse_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_tse_adr_w-1 DOWNTO 0),
      avs_eth_0_tse_write_export                => eth1g_tse_mosi.wr,
      avs_eth_0_tse_read_export                 => eth1g_tse_mosi.rd,
      avs_eth_0_tse_writedata_export            => eth1g_tse_mosi.wrdata(c_word_w-1 DOWNTO 0),
      avs_eth_0_tse_readdata_export             => eth1g_tse_miso.rddata(c_word_w-1 DOWNTO 0),
      avs_eth_0_tse_waitrequest_export          => eth1g_tse_miso.waitrequest,
      avs_eth_0_reg_address_export              => eth1g_reg_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_eth_adr_w-1 DOWNTO 0),
      avs_eth_0_reg_write_export                => eth1g_reg_mosi.wr,
      avs_eth_0_reg_read_export                 => eth1g_reg_mosi.rd,
      avs_eth_0_reg_writedata_export            => eth1g_reg_mosi.wrdata(c_word_w-1 DOWNTO 0),
      avs_eth_0_reg_readdata_export             => eth1g_reg_miso.rddata(c_word_w-1 DOWNTO 0),
      avs_eth_0_ram_address_export              => eth1g_ram_mosi.address(c_unb2b_board_peripherals_mm_reg_default.ram_eth_adr_w-1 DOWNTO 0),
      avs_eth_0_ram_write_export                => eth1g_ram_mosi.wr,
      avs_eth_0_ram_read_export                 => eth1g_ram_mosi.rd,
      avs_eth_0_ram_writedata_export            => eth1g_ram_mosi.wrdata(c_word_w-1 DOWNTO 0),
      avs_eth_0_ram_readdata_export             => eth1g_ram_miso.rddata(c_word_w-1 DOWNTO 0),
      avs_eth_0_irq_export                      => eth1g_reg_interrupt,

      reg_unb_sens_address_export               => reg_unb_sens_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_unb_sens_adr_w-1 DOWNTO 0),
      reg_unb_sens_write_export                 => reg_unb_sens_mosi.wr,
      reg_unb_sens_writedata_export             => reg_unb_sens_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_unb_sens_read_export                  => reg_unb_sens_mosi.rd,
      reg_unb_sens_readdata_export              => reg_unb_sens_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_unb_pmbus_address_export              => reg_unb_pmbus_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_unb_pmbus_adr_w-1 DOWNTO 0),
      reg_unb_pmbus_write_export                => reg_unb_pmbus_mosi.wr,
      reg_unb_pmbus_writedata_export            => reg_unb_pmbus_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_unb_pmbus_read_export                 => reg_unb_pmbus_mosi.rd,
      reg_unb_pmbus_readdata_export             => reg_unb_pmbus_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_fpga_temp_sens_address_export         => reg_fpga_temp_sens_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_fpga_temp_sens_adr_w-1 DOWNTO 0),
      reg_fpga_temp_sens_write_export           => reg_fpga_temp_sens_mosi.wr,
      reg_fpga_temp_sens_writedata_export       => reg_fpga_temp_sens_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_fpga_temp_sens_read_export            => reg_fpga_temp_sens_mosi.rd,
      reg_fpga_temp_sens_readdata_export        => reg_fpga_temp_sens_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_fpga_voltage_sens_address_export      => reg_fpga_voltage_sens_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_fpga_voltage_sens_adr_w-1 DOWNTO 0),
      reg_fpga_voltage_sens_write_export        => reg_fpga_voltage_sens_mosi.wr,
      reg_fpga_voltage_sens_writedata_export    => reg_fpga_voltage_sens_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_fpga_voltage_sens_read_export         => reg_fpga_voltage_sens_mosi.rd,
      reg_fpga_voltage_sens_readdata_export     => reg_fpga_voltage_sens_miso.rddata(c_word_w-1 DOWNTO 0),

      rom_system_info_address_export            => rom_unb_system_info_mosi.address(c_unb2b_board_peripherals_mm_reg_default.rom_unb_system_info_adr_w-1 DOWNTO 0), 
      rom_system_info_write_export              => rom_unb_system_info_mosi.wr,
      rom_system_info_writedata_export          => rom_unb_system_info_mosi.wrdata(c_word_w-1 DOWNTO 0),
      rom_system_info_read_export               => rom_unb_system_info_mosi.rd,
      rom_system_info_readdata_export           => rom_unb_system_info_miso.rddata(c_word_w-1 DOWNTO 0),

      pio_system_info_address_export            => reg_unb_system_info_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_unb_system_info_adr_w-1 DOWNTO 0), 
      pio_system_info_write_export              => reg_unb_system_info_mosi.wr,
      pio_system_info_writedata_export          => reg_unb_system_info_mosi.wrdata(c_word_w-1 DOWNTO 0),
      pio_system_info_read_export               => reg_unb_system_info_mosi.rd,
      pio_system_info_readdata_export           => reg_unb_system_info_miso.rddata(c_word_w-1 DOWNTO 0),

      pio_pps_address_export                    => reg_ppsh_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_ppsh_adr_w-1 DOWNTO 0),
      pio_pps_write_export                      => reg_ppsh_mosi.wr,
      pio_pps_writedata_export                  => reg_ppsh_mosi.wrdata(c_word_w-1 DOWNTO 0),
      pio_pps_read_export                       => reg_ppsh_mosi.rd,
      pio_pps_readdata_export                   => reg_ppsh_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_wdi_address_export                    => reg_wdi_mosi.address(0 DOWNTO 0),
      reg_wdi_write_export                      => reg_wdi_mosi.wr,
      reg_wdi_writedata_export                  => reg_wdi_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_wdi_read_export                       => reg_wdi_mosi.rd,
      reg_wdi_readdata_export                   => reg_wdi_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_remu_address_export                   => reg_remu_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_remu_adr_w-1 DOWNTO 0),
      reg_remu_write_export                     => reg_remu_mosi.wr,
      reg_remu_writedata_export                 => reg_remu_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_remu_read_export                      => reg_remu_mosi.rd,
      reg_remu_readdata_export                  => reg_remu_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_epcs_address_export                   => reg_epcs_mosi.address(c_unb2b_board_peripherals_mm_reg_default.reg_epcs_adr_w-1 DOWNTO 0),
      reg_epcs_write_export                     => reg_epcs_mosi.wr,
      reg_epcs_writedata_export                 => reg_epcs_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_epcs_read_export                      => reg_epcs_mosi.rd,
      reg_epcs_readdata_export                  => reg_epcs_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_dpmm_ctrl_address_export              => reg_dpmm_ctrl_mosi.address(0 DOWNTO 0),
      reg_dpmm_ctrl_write_export                => reg_dpmm_ctrl_mosi.wr,
      reg_dpmm_ctrl_writedata_export            => reg_dpmm_ctrl_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_dpmm_ctrl_read_export                 => reg_dpmm_ctrl_mosi.rd,
      reg_dpmm_ctrl_readdata_export             => reg_dpmm_ctrl_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_mmdp_data_address_export              => reg_mmdp_data_mosi.address(0 DOWNTO 0),
      reg_mmdp_data_write_export                => reg_mmdp_data_mosi.wr,
      reg_mmdp_data_writedata_export            => reg_mmdp_data_mosi.wrdata(c_word_w-1 DOWNTO 0),
      reg_mmdp_data_read_export                 => reg_mmdp_data_mosi.rd,
      reg_mmdp_data_readdata_export             => reg_mmdp_data_miso.rddata(c_word_w-1 DOWNTO 0),

      reg_dpmm_data_address_export              => reg_dpmm_data_mosi.address(0 DOWNTO 0),
      reg_dpmm_data_read_export                 => reg_dpmm_data_mosi.rd,
      reg_dpmm_data_readdata_export             => reg_dpmm_data_miso.rddata(c_word_w-1 DOWNTO 0),
      reg_dpmm_data_write_export                => reg_dpmm_data_mosi.wr,
      reg_dpmm_data_writedata_export            => reg_dpmm_data_mosi.wrdata(c_word_w-1 DOWNTO 0),

      reg_mmdp_ctrl_address_export              => reg_mmdp_ctrl_mosi.address(0 DOWNTO 0),
      reg_mmdp_ctrl_read_export                 => reg_mmdp_ctrl_mosi.rd,
      reg_mmdp_ctrl_readdata_export             => reg_mmdp_ctrl_miso.rddata(c_word_w-1 DOWNTO 0),
      reg_mmdp_ctrl_write_export                => reg_mmdp_ctrl_mosi.wr,
      reg_mmdp_ctrl_writedata_export            => reg_mmdp_ctrl_mosi.wrdata(c_word_w-1 DOWNTO 0),

      reg_ta2_unb2b_jesd204b_address_export     => reg_ta2_unb2b_jesd204b_mosi.address(7 DOWNTO 0),  
      reg_ta2_unb2b_jesd204b_read_export        => reg_ta2_unb2b_jesd204b_mosi.rd, 
      reg_ta2_unb2b_jesd204b_readdata_export    => reg_ta2_unb2b_jesd204b_miso.rddata(c_word_w-1 DOWNTO 0), 
      reg_ta2_unb2b_jesd204b_write_export       => reg_ta2_unb2b_jesd204b_mosi.wr, 
      reg_ta2_unb2b_jesd204b_writedata_export   => reg_ta2_unb2b_jesd204b_mosi.wrdata(c_word_w-1 DOWNTO 0), 
      reg_ta2_unb2b_jesd204b_waitrequest_export => reg_ta2_unb2b_jesd204b_miso.waitrequest, 

      kernel_cra_waitrequest                    => board_kernel_cra_waitrequest,            
      kernel_cra_readdata                       => board_kernel_cra_readdata,               
      kernel_cra_readdatavalid                  => board_kernel_cra_readdatavalid,          
      kernel_cra_burstcount                     => board_kernel_cra_burstcount,              
      kernel_cra_writedata                      => board_kernel_cra_writedata,               
      kernel_cra_address                        => board_kernel_cra_address,                 
      kernel_cra_write                          => board_kernel_cra_write,                   
      kernel_cra_read                           => board_kernel_cra_read,                    
      kernel_cra_byteenable                     => board_kernel_cra_byteenable,              
      kernel_cra_debugaccess                    => board_kernel_cra_debugaccess,    
         
      kernel_irq_irq                            => board_kernel_irq_irq,            

      kernel_register_mem_address               => board_kernel_register_mem_address,
      kernel_register_mem_clken                 => board_kernel_register_mem_clken, 
      kernel_register_mem_chipselect            => board_kernel_register_mem_chipselect,
      kernel_register_mem_write                 => board_kernel_register_mem_write, 
      kernel_register_mem_readdata              => board_kernel_register_mem_readdata,
      kernel_register_mem_writedata             => board_kernel_register_mem_writedata,
      kernel_register_mem_byteenable            => board_kernel_register_mem_byteenable        

  );

END str;

